from setuptools import setup, find_packages

setup(
    name="noyerTransformer",
    version="0.1.0",
    author="Elowarp",
    author_email="elowarp@gmail.com",
    description="Un convertisseur de fichier .tex en .md pour Mkdocs",
    license="MIT",
    packages=find_packages(),
    url="https://github.com/Elowarp/latexToMkdocs"
)